--[[
Name: LibCooldown-Item-1.0
Revision: $Revision: 53235 $
Author(s): Sirow (EU-Gul'dan)
Documentation: http://www.wowace.com/wiki/LibCooldown-1.0
SVN: http://svn.wowace.com/wowace/trunk/LibCooldown-1.0/
Description: Register item cooldowns to descriptive events
Dependencies: AceLibrary, AceOO-2.0, AceEvent-2.0
]]


-------------------------------------------------------------------------------
-- Library - Intro
-------------------------------------------------------------------------------
-- Major and minor version
local vmajor = "LibCooldown-Item-1.0"
local vminor = "$Revision: 53235 $"


-- Make sure require Libs/Framework is loaded
if (not AceLibrary) then error(vmajor .. " requires <AceLibrary>") end
if (not AceLibrary:IsNewVersion(vmajor, vminor)) then return end
if (not AceLibrary:HasInstance("AceEvent-2.0")) then error(vmajor .. " requires <AceEvent-2.0>") end


-- This table gonna be a library. Once upon a time...
local lib = {}


-- Embed/Delegate
local AceEvent 	= AceLibrary("AceEvent-2.0")
AceEvent:embed(lib)


-- Activate a new instance of this library
local function activate(self, oldLib, oldDeactivate)
	if oldLib then 
		self.vars = oldLib.vars

		oldLib:UnregisterEvent("BAG_UPDATE_COOLDOWN")
		oldLib:UnregisterEvent("UNIT_INVENTORY_CHANGED")
		
		oldLib:DeleteAllItemCooldowns()
	else
		self.vars = {}
	end
	
	-- Items in this table get monitored
	self.vars.itemCache = {}
	-- Items in this table get monitored (this are equipped items
	self.vars.eItemCache = {}
	
	-- This table holds all registered items
	self.vars.registeredItems = {}
	
	-- Register Events
	self:RegisterEvent("BAG_UPDATE_COOLDOWN")
	self:RegisterEvent("UNIT_INVENTORY_CHANGED")

	if oldDeactivate then oldDeactivate(oldLib) end
end


-------------------------------------------------------------------------------
-- Library - Locals
-------------------------------------------------------------------------------
-- Possible slot names (equipment slots)
local slotNames = { 
	"HeadSlot", 
	"NeckSlot", 
	"ShoulderSlot", 
	"BackSlot", 
	"ChestSlot", 
	"ShirtSlot", 
	"TabardSlot", 
	"WristSlot", 
	"HandsSlot", 
	"WaistSlot", 
	"LegsSlot", 
	"FeetSlot", 
	"Finger0Slot",
	"Finger1Slot",  
	"Trinket0Slot", 
	"Trinket1Slot", 
	"MainHandSlot", 
	"SecondaryHandSlot", 
	"RangedSlot", 
	"AmmoSlot",
} 

-- Convert from GetItemInfo to slotID(s)
local equipLocToSlot = {
	["INVTYPE_AMMO"] = { 0 },
	["INVTYPE_HEAD"] = { 1 },
	["INVTYPE_NECK"] = { 2 },
	["INVTYPE_SHOULDER"] = { 3 },
	["INVTYPE_BODY"] = { 4 },
	["INVTYPE_CHEST"] = { 5 },
	["INVTYPE_ROBE"] = { 5 },
	["INVTYPE_WAIST"] = { 6 },
	["INVTYPE_LEGS"] = { 7 },
	["INVTYPE_FEET"] = { 8 },
	["INVTYPE_WRIST"] = { 9 },
	["INVTYPE_HAND"] = { 10 },
	["INVTYPE_FINGER"] = { 11, 12 },
	["INVTYPE_TRINKET"] = { 13, 14 },
	["INVTYPE_CLOAK"] = { 15 },
	["INVTYPE_WEAPON"] = { 16, 17 },
	["INVTYPE_SHIELD"] = { 17 },
	["INVTYPE_2HWEAPON"] = { 16 },
	["INVTYPE_WEAPONMAINHAND"] = { 16 },
	["INVTYPE_WEAPONOFFHAND"] = { 17 },
	["INVTYPE_HOLDABLE"] = { 17 },
	["INVTYPE_RANGED"] = { 18 },
	["INVTYPE_THROWN"] = { 18 },
	["INVTYPE_RANGEDRIGHT"] = { 18 },
	["INVTYPE_RELIC"] = { 18 },
	["INVTYPE_TABARD"] = { 19 },
	["INVTYPE_BAG"] = { 20, 21, 22, 23 },
	["INVTYPE_QUIVER"] = { 20, 21, 22, 23 },
}


-------------------------------------------------------------------------------
-- Library - Core
-------------------------------------------------------------------------------
function lib:UNIT_INVENTORY_CHANGED(unit)
	if (unit == "player") then
		-- Rescan position of itemName/itemID items
		self:ScanItems()
		
		-- Rescan position of equipped items
		self:ScanEquippedItems()
	end
end
function lib:BAG_UPDATE_COOLDOWN()
	-- Cooldowns by itemName/itemID
	self:UpdateItemCooldown()
	
	-- Cooldowns by equipment
	self:UpdateEquippedItemCooldown()
end


-- Scan all monitored items
function lib:ScanItems()
	-- Scan all monitored items
	for itemID, _ in pairs(self.vars.registeredItems) do
		self:ScanItem(itemID)
	end
end


-- Scan for a certain item
function lib:ScanItem(itemID)
	local found
	
	-- Scan last bag position
	found = self:ScanItem_LastBagLoc(itemID)

	-- Scan equipment
	if (not found) then
		found = self:ScanItem_Equipped(itemID)
	end
	
	-- Scan all bags
	if (not found) then
		found = self:ScanItem_Bag(itemID)
	end

	-- Still not found? (Well then the item is gone)
	if (not found) then
		-- Remove cooldown trigger
		if (self:IsEventScheduled(string.format("LibCD_ItemCooldownShedule_%d", itemID))) then
			self:CancelScheduledEvent(string.format("LibCD_ItemCooldownShedule_%d", itemID))
		end
		
		-- Clear item from beeing watched, since we couldn't find it
		self.vars.itemCache[itemID] = nil
	end
	
	return found
end


-- Scan (known) bag slot for this item (its the slot the item USED to be)
function lib:ScanItem_LastBagLoc(itemID)
	-- See if we allready know where the item is located
	if (self.vars.itemCache[itemID] and self.vars.itemCache[itemID].LastBagLoc) then
		-- Get ItemLink
		local itemLink = GetContainerItemLink(self.vars.itemCache[itemID].LastBagLoc.bagID, self.vars.itemCache[itemID].LastBagLoc.slotID)
		
		-- Compare item we are searching for with item on this (known) slot
		if (itemLink and self.vars.registeredItems[itemID] == GetItemInfo(itemLink)) then
			self:UnequipItem(itemID)
			return true
			
		-- Not at the known location anymore
		else
			return false
		end
	end
	
	return false
end


-- Scan equipment for this item (IF it is equipable)
function lib:ScanItem_Equipped(itemID)	
	-- If item is equipable, search equipment for item
	if (IsEquippableItem(itemID)) then
		-- Were is it equipable?
		local itemEquipLoc, itemLink
		_, _, _, _, _, _, _, _, itemEquipLoc, _ = GetItemInfo(itemID)
	
		-- For each equipable position test if item can be found there (max two locations)
		for _, slotID in pairs(equipLocToSlot[itemEquipLoc]) do
			itemLink = GetInventoryItemLink("player", slotID)
			
			-- Compare name of equipped item with name of item we are searching for
			if (itemLink and self.vars.registeredItems[itemID] == GetItemInfo(itemLink)) then
				-- Create new entry or update old one
				if (self.vars.itemCache[itemID]) then
					-- Test if item was equipped to another slot
					if (self.vars.itemCache[itemID].LastEquipLoc ~= slotID) then
						self:UnequipItem(itemID)
					end
					
					self.vars.itemCache[itemID].isEquipped = true
					self.vars.itemCache[itemID].LastEquipLoc =  slotID
				else
					self.vars.itemCache[itemID] = {
						isEquipped = true,
						LastEquipLoc =  slotID,
						startTime = 0
					}
				end
			
				return true
			end
		end
		
		self:UnequipItem(itemID)
	end
	
	return false
end


-- (Re-)scan complete inventory for this item
function lib:ScanItem_Bag(itemID)
	local itemLink
	
	-- Search through ALL bags
	for bagID = 0, 4 do
		for slotID = 1, GetContainerNumSlots(bagID) do
			-- Item-link
			itemLink = GetContainerItemLink(bagID, slotID)
			
			-- Compare this item with the one we're searching for
			if (itemLink and self.vars.registeredItems[itemID] == GetItemInfo(itemLink)) then
				-- Create new entry or update old one
				if (self.vars.itemCache[itemID]) then
					self.vars.itemCache[itemID].LastBagLoc = {
						bagID = bagID,
						slotID = slotID
					}
					self:UnequipItem(itemID)
				else
					self.vars.itemCache[itemID] = {
						LastBagLoc = {
							bagID = bagID,
							slotID = slotID
						},
						isEquipped = false,
						startTime = 0
					}
				end
				
				return true
			end
		end
	end
	
	return false
end


-- Item isn't equipped, if it was equipped, cancel Sheduled event
function lib:UnequipItem(itemID)
	if (self.vars.itemCache[itemID] and self.vars.itemCache[itemID].isEquipped) then
		self.vars.itemCache[itemID].isEquipped = false
		
		if (self:IsEventScheduled(string.format("LibCD_ItemCooldownShedule_%d", itemID))) then
			self:CancelScheduledEvent(string.format("LibCD_ItemCooldownShedule_%d", itemID))
		end
	end
end	


-- Update all item cooldowns
function lib:UpdateItemCooldown()
	-- Temp variables
	local startTime, duration, hasCooldown
	
	-- For each item that should be monitored
	for itemID, itemInfo in pairs(self.vars.itemCache) do
		-- Make sure we know were the item is located 
		-- (Unless item was moved this should be in O(1))
		if (self:ScanItem(itemID)) then
			-- Scan cooldown of equipped item
			if (itemInfo.isEquipped) then
				startTime, duration, hasCooldown = GetInventoryItemCooldown("player", itemInfo.LastEquipLoc)
			-- Scan cooldown of bags item
			else
				startTime, duration, hasCooldown = GetContainerItemCooldown(itemInfo.LastBagLoc.bagID, itemInfo.LastBagLoc.slotID)
			end
		
			-- Does this item have a cooldown? (Unless someone added item without CD, this should't happen)
			if (hasCooldown == 1) then
				-- Lets see if a NEW cooldown was activated
				if (itemInfo.startTime ~= startTime) then
					-- Update start-time
					itemInfo.startTime = startTime
					
					-- Trigger differently
					if (itemInfo.isEquipped) then				
						-- Shedule event with duration and trigger event
						self:TriggerEvent(
							"LibCD_ItemCooldownStart", 
							self.vars.registeredItems[itemID],
							itemID,
							duration,
							nil,
							itemInfo.LastEquipLoc
						)
						
						self:ScheduleEvent(
							string.format("LibCD_ItemCooldownShedule_%d", itemID), 
							self.TriggerEvent, 
							duration, 
							self, 
							"LibCD_ItemCooldownEnd", 
							self.vars.registeredItems[itemID],
							itemID,
							startTime,
							nil,
							itemInfo.LastEquipLoc
						)
					else
						-- Shedule event with duration and trigger event
						self:TriggerEvent(
							"LibCD_ItemCooldownStart", 
							self.vars.registeredItems[itemID],
							itemID,
							duration,
							itemInfo.LastBagLoc.bagID,
							itemInfo.LastBagLoc.slotID
						)
						
						self:ScheduleEvent(
							string.format("LibCD_ItemCooldownShedule_%d", itemID), 
							self.TriggerEvent, 
							duration, 
							self, 
							"LibCD_ItemCooldownEnd", 
							self.vars.registeredItems[itemID],
							itemID,
							startTime,
							itemInfo.LastBagLoc.bagID,
							itemInfo.LastBagLoc.slotID
						)
					end
				end
			end
		end
	end
end


-- Scan if equipment has changed (complete character slots)
function lib:ScanEquippedItems()
	local slotID
	
	for _, slotName in pairs(slotNames) do
		slotID = GetInventorySlotInfo(slotName)
	
		self:ScanEquippedItem(slotID)
	end
end


-- Scan if equipment has changed
function lib:ScanEquippedItem(slotID)
	-- Last item on this slot kown?
	if (self.vars.eItemCache[slotID]) then
		itemLink = GetInventoryItemLink("player", slotID)
		
		-- Is item equipped?
		if (itemLink) then
			-- Compare name of equipped item with name of item that was last (known) equipped
			if (self.vars.eItemCache[slotID].LastItemName ~= GetItemInfo(itemLink)) then
				self.vars.eItemCache[slotID] = {
					startTime = 0,
					LastItemName = GetItemInfo(itemLink)
				}
				
				-- Delete stop trigger, since item was exchanged
				if (self:IsEventScheduled(string.format("LibCD_EquippedItemCooldownShedule_%d", slotID))) then
					self:CancelScheduledEvent(string.format("LibCD_EquippedItemCooldownShedule_%d", slotID))
				end
					
				return true
				
			-- Still the same equipped
			else
				return true
			end
		
		-- Item unequipped (Item WAS equipped)
		else
			self.vars.eItemCache[slotID] = nil
			
			-- Delete stop trigger, si nce item isn't equipped anymore
			if (self:IsEventScheduled(string.format("LibCD_EquippedItemCooldownShedule_%d", slotID))) then
				self:CancelScheduledEvent(string.format("LibCD_EquippedItemCooldownShedule_%d", slotID))
			end
			
			return false
		end
		
	-- No known item on this slot, scan for new item
	else
		itemLink = GetInventoryItemLink("player", slotID)
		
		-- Is item equipped
		if (itemLink) then
			self.vars.eItemCache[slotID] = {
				startTime = 0,
				LastItemName = GetItemInfo(itemLink)
			}
			
			return true
			
		-- No item equipped
		else
			return false
		end
	end
end


-- Scan equipped items for updated cooldowns
function lib:UpdateEquippedItemCooldown()
	-- Temp variables
	local startTime, duration, hasCooldown, itemID
	
	-- For each item that should be monitored
	for slotID, itemInfo in pairs(self.vars.eItemCache) do
		if (self:ScanEquippedItem(slotID)) then
			-- Get cooldown info
			startTime, duration, hasCooldown = GetInventoryItemCooldown("player", slotID)
			
			-- Does this item have a cooldown? (Unless someone added item without CD, this should't happen)
			if (hasCooldown == 1) then
				-- Lets see if a NEW cooldown was activated
				if (self.vars.eItemCache[slotID].startTime ~= startTime) then
					-- Update start-time
					self.vars.eItemCache[slotID].startTime = startTime
					
					-- Fetch itemID
					itemID = GetInventoryItemLink("player", slotID):match("item:(%d+):")
					
					-- Shedule event with duration and trigger event
					self:TriggerEvent(
						"LibCD_EquippedItemCooldownStart", 
						self.vars.eItemCache[slotID].itemName,
						itemID,
						duration,
						slotID
					)
					
					self:ScheduleEvent(
						string.format("LibCD_EquippedItemCooldownShedule_%d", itemID), 
						self.TriggerEvent, 
						duration, 
						self, 
						"LibCD_EquippedItemCooldownEnd", 
						self.vars.eItemCache[slotID].itemName,
						itemID,
						startTime,
						slotID
					)
				end
			end
		end
	end
end


-------------------------------------------------------------------------------
-- Library - API
-------------------------------------------------------------------------------
-- Add new spell to be listened to
function lib:AddItemCoodown(itemNameOrID)
	local itemName, itemLink = GetItemInfo(itemNameOrID) 
	
	-- If item found/exists, add it
	if (itemName ~= nil) then
		-- find itemID
		itemID = itemLink:match ("item:(%d+):")
	
		if (itemID ~= nil) then
			if (not self.vars.registeredItems[itemID]) then
				-- Add to monitor and rescan everything
				self.vars.registeredItems[itemID] = itemName
				
				self:ScanItem(itemID)
			end
			
			return true
		else
			return false
		end
	else
		return false
	end
end


-- Delete spell that shouldn't be listened to anymore
function lib:DeleteItemCooldown(itemNameOrID)
	local itemLink
	_, itemLink = GetItemInfo(itemNameOrID) 
	
	-- If item found/exists, add it
	if (itemLink ~= nil) then
		-- find itemID
		itemID = itemLink:match ("item:(%d+):")
	
		if (itemID ~= nil) then
			-- Unregister sheduled event
			if (self:IsEventScheduled("LibCD_ItemCooldownShedule_" .. itemID)) then
				self:CancelScheduledEvent("LibCD_ItemCooldownShedule_" .. itemID)
			end

			self.vars.registeredItems[itemID] = nil
			self.vars.itemCache[itemID] = nil
		end
	end
end


-- Listen to all spells of own Spellbook
function lib:AddAllItemCooldowns()
	local itemName, startTime, duration, hasCooldown
	
	-- View all bag items
	for bagID = 0, 4 do
		for slotID = 1, GetContainerNumSlots(bagID) do
			_, _, hasCooldown = GetContainerItemCooldown(bagID, slotID)

			-- Item available?
			if (hasCooldown == 1 and GetContainerItemLink(bagID, slotID)) then
				self:AddItemCoodown(GetContainerItemLink(bagID, slotID))
			end
		end
	end
	
	-- View all equipped items
	local slotID_
	for _, slotName in pairs(slotNames) do
		slotID_ = GetInventorySlotInfo(slotName)
		_, _, hasCooldown = GetInventoryItemCooldown("player", slotID_)
		
		if (hasCooldown  == 1 and GetInventoryItemLink("player", slotID_)) then
			self:AddItemCoodown(GetInventoryItemLink("player", slotID_))
		end
	end
end


-- Delete all listener
function lib:DeleteAllItemCooldowns()
	for spellName, _ in pairs(self.vars.registeredItems) do
		self:DeleteItemCooldown(itemNameOrID)
	end
end


-------------------------------------------------------------------------------
-- Library - Outro
-------------------------------------------------------------------------------
AceLibrary:Register(lib, vmajor, vminor, activate)
lib = nil
