--[[
Name: LibCooldown-Spell-1.0
Revision: $Revision: 53235 $
Author(s): Sirow (EU-Gul'dan)
Documentation: http://www.wowace.com/wiki/LibCooldown-1.0
SVN: http://svn.wowace.com/wowace/trunk/LibCooldown-1.0/
Description: Register spell cooldowns to descriptive events
Dependencies: AceLibrary, AceOO-2.0, AceEvent-2.0, Gratuity
]]


-------------------------------------------------------------------------------
-- Library - Intro
-------------------------------------------------------------------------------
-- Major and minor version
local vmajor = "LibCooldown-Spell-1.0"
local vminor = "$Revision: 53235 $"


-- Make sure require Libs/Framework is loaded
if (not AceLibrary) then error(vmajor .. " requires <AceLibrary>") end
if (not AceLibrary:IsNewVersion(vmajor, vminor)) then return end
if (not AceLibrary:HasInstance("AceEvent-2.0")) then error(vmajor .. " requires <AceEvent-2.0>") end
if (not AceLibrary:HasInstance("Gratuity-2.0")) then error(vmajor .. " requires <Gratuity-2.0>") end


-- This table gonna be a library. Once upona time...
local lib = {}


-- Embed/Delegate
local AceEvent 	= AceLibrary("AceEvent-2.0")
local Gratuity = AceLibrary("Gratuity-2.0")
AceEvent:embed(lib)


-- Activate a new instance of this library
local function activate(self, oldLib, oldDeactivate)
	if oldLib then 
		self.vars = oldLib.vars

		oldLib:UnregisterEvent("SPELLS_CHANGED")
		oldLib:UnregisterEvent("SPELL_UPDATE_COOLDOWN")
		
		oldLib:DeleteAllSpellCooldowns()
	else
		self.vars = {}
	end
	
	-- Spells in this table get monitored
	self.vars.spellCache = {}


	-- This table holds all registered spells
	self.vars.registeredSpells = {}
	
	-- Register Events
	self:RegisterEvent("SPELL_UPDATE_COOLDOWN")
	self:RegisterEvent("SPELLS_CHANGED")

	if oldDeactivate then oldDeactivate(oldLib) end
end


-------------------------------------------------------------------------------
-- Library - Core
-------------------------------------------------------------------------------
function lib:SPELLS_CHANGED()
	self:ScanSpells()
end
function lib:SPELL_UPDATE_COOLDOWN()
	self:UpdateSpellCooldown()
end


-- Scan Speelbook for spells that have a cooldown
function lib:ScanSpells()
	-- Pre-loop stuff
	local i = 1
	local previousspellName, scanSpell
	local spellName = GetSpellName(i, BOOKTYPE_SPELL)
	
	-- Scan through every spell in spellbook
	while spellName do
		-- Prevent rescanning of spells with the same name (lets assume ALL have the same CD!)
		-- And only monitor those spells that are told be be listened to
		if (spellName ~= previousspellName and self.vars.registeredSpells[spellName]) then
			-- Scan spell tooltip with gratuity to find cooldown
			Gratuity:SetSpell(i, BOOKTYPE_SPELL)
			
			-- Found cooldown text?
			if (Gratuity:Find("cooldown", 2, 3, true)) then
				self.vars.spellCache[spellName] = {
					spellID = i,
					startTime = 0,
				}
				
			-- For some reason, this spell lost his cooldown?!
			elseif (self.vars.spellCache[spellName]) then
				self.vars.spellCache[spellName] = nil
			end
		end
	
		-- Set values for next loop
		i = i + 1
		previousspellName = spellName
		spellName = GetSpellName(i, BOOKTYPE_SPELL)
	end
	
	-- Also scan if any cooldown updated
	self:UpdateSpellCooldown()
end


function lib:SpellFound(spellName)
	-- Test if spell is still at this spellID
	if (GetSpellName(self.vars.spellCache[spellName].spellID, BOOKTYPE_SPELL) == spellName) then
		return true
	else
		self:ScanSpells()
		
		-- Spell found or not?
		if (self.vars.spellCache[spellName]) then
			return true
		else
			return false
		end
	end
end


-- Update all item cooldowns
function lib:UpdateSpellCooldown()
	local startTime, duration, enable
	
	-- For each cache/saved spell...
	for spellName, spellInfo in pairs(self.vars.spellCache) do
		-- Make sure spell still available (Should be still in O(1)!)
		if (self:SpellFound(spellName)) then
			-- ... get spell info
			startTime, duration, enable = GetSpellCooldown(spellInfo.spellID, BOOKTYPE_SPELL)
			
			-- if cooldown is started, start animation
			if (enable == 1 and spellInfo.startTime ~= startTime) then
				-- Update startTime
				spellInfo.startTime = startTime
				
				-- Shedule event with duration and trigger event
				self:TriggerEvent(
					"LibCD_SpellCooldownStart", 
					spellName,
					spellInfo.spellID,
					duration
				)
				
				self:ScheduleEvent(
					"LibCD_SpellCooldownShedule_" .. spellName, 
					self.TriggerEvent, 
					duration, 
					self, 
					"LibCD_SpellCooldownEnd", 
					spellName,
					spellInfo.spellID,
					spellInfo.startTime
				)
			end
		end
	end
end


-------------------------------------------------------------------------------
-- Library - API
-------------------------------------------------------------------------------
-- Add new spell to be listened to
function lib:AddSpellCoodown(spellName)
	if (not self.vars.registeredSpells[spellName]) then
		self.vars.registeredSpells[spellName] = true
		self:ScanSpells()
		
		return true
	else
		return false
	end
end


-- Delete spell that shouldn't be listened to anymore
function lib:DeleteSpellCooldown(spellName)
	-- Unregister sheduled event
	if (self:IsEventScheduled("LibCD_SpellCooldownShedule_" .. spellName)) then
		self:CancelScheduledEvent("LibCD_SpellCooldownShedule_" .. spellName)
	end

	self.vars.registeredSpells[spellName] = nil
	self.vars.spellCache[spellName] = nil
end


-- Listen to all spells of own Spellbook
function lib:AddAllSpellCooldowns()
	-- Pre-loop stuff
	local i = 1
	local previousspellName, scanSpell
	local spellName = GetSpellName(i, BOOKTYPE_SPELL)
	
	-- Scan through every spell in spellbook
	while spellName do
		if (spellName ~= previousspellName) then
			self:AddSpellCoodown(spellName)
		end
	
		-- Set values for next loop
		i = i + 1
		previousspellName = spellName
		spellName = GetSpellName(i, BOOKTYPE_SPELL)
	end
end


-- Delete all listener
function lib:DeleteAllSpellCooldowns()
	for spellName, _ in pairs(self.vars.registeredSpells) do
		self:DeleteSpellCooldown(spellName)
	end
end


-------------------------------------------------------------------------------
-- Library - Outro
-------------------------------------------------------------------------------
AceLibrary:Register(lib, vmajor, vminor, activate)
lib = nil
